﻿/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|          Program Name: WinLib                                                                                         |
|                Author: Michael Krämer                                                                                 |
|                        Michael@dk8pp.de                                                                               |
|                  Date: 03-Dec-2014                                                                                    |
|              Language: Visual C Express 2013                                                                          |
|               Version: 0.3                                                                                            |
|        Program status: Refurbished for Visual C, support for 64-bit and Unicode                                       |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Library with some often used functions for Windows programming                                              |
|                                                                                                                       |
|           Define _CONSOLE for character mode programs                                                                 |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdio.h>
#include <windows.h>
#include <commctrl.h>
#include <stdlib.h>
#include <malloc.h>
#include <errno.h>
#include <tchar.h>
#include "resource.h"
#include "WinLib.h"

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define STRINGSIZE 256                  /* default size for strings in number of characters */
#define DYN_BUFFERCHARSIZE (4*1024)     /* Size of dynamically allocated buffers in number of characters.
                                           Don't make it excessively long, because sprintf_s regularly clears
                                           the buffer to the end */
#define DEFAULT_SNAP_X 20               /* Default SNAP range when moving window horizontally */
#define DEFAULT_SNAP_Y 20               /* Default SNAP range when moving window vertically */
#define SKN_LENGTH 100                  /* length of sub key name (take care! allocted on stack) */
#define ESCAPE_CHAR _T('%')             /* for conversion of multi size strings */
#define MULTI_SZ_CHAR _T('_')           /* for conversion of multi size strings */

#define EMAILADDR _T("Michael@dk8pp.de")
#define WEBSITE _T("http://freeware.dk8pp.de/")
#ifdef _CONSOLE
#define PROGRAMMER _T("Michael Kr„mer")
#define NEWLINE _T("\n")
#else
#define PROGRAMMER _T("Michael Krämer")
#define NEWLINE _T("\r\r\n")
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       external variables                                                                                              |
+-----------------------------------------------------------------------------------------------------------------------+
*/
extern TCHAR programname[] ;
extern int programversion ;
extern HWND      MainWindow ;
extern HINSTANCE MyInstance ;
extern CLO clo[] ;
extern size_t clo_size ;
extern TCHAR purpose[] ;

#ifndef _CONSOLE
/* The following externs are only defined in windowed mode */
extern COORDS coordinates[] ;
extern COORDS *sorted_coord[] ;
extern size_t coord_size ;
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/
static _TCHAR Info_Strng[STRINGSIZE] ;
static int process_prio[] = {IDLE_PRIORITY_CLASS, NORMAL_PRIORITY_CLASS, HIGH_PRIORITY_CLASS, REALTIME_PRIORITY_CLASS} ;
static int thread_prio[] = {THREAD_PRIORITY_IDLE, THREAD_PRIORITY_LOWEST, THREAD_PRIORITY_BELOW_NORMAL,
                            THREAD_PRIORITY_NORMAL, THREAD_PRIORITY_ABOVE_NORMAL, THREAD_PRIORITY_HIGHEST,
                            THREAD_PRIORITY_TIME_CRITICAL} ;
static uint32_t helpdlgitem ;
static TCHAR inifilename[MAX_PATH+1] ;
static TCHAR local_strng1[STRINGSIZE] ;
static TCHAR help_header[STRINGSIZE] ;
static TCHAR HelpWindowName[STRINGSIZE] = _T("") ;
static uint32_t GetHelp_pos ;
static TCHAR errorparm[MAXPARMLENGTH] ;

#ifndef _CONSOLE
static HWND HelpWindow ;
static int32_t min_size_x ;     /* Minimum horizontal dialog size */
static int32_t min_size_y ;     /* Minimum vertical dialog size */
int32_t snap_x, snap_y ;        /* horizontal and vertical snap distances */
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/
INT_PTR CALLBACK DlgHelp(HWND window_handle, unsigned msg, WPARAM wparam, LPARAM lparam) ;
static INT_PTR CALLBACK DlgInfo(HWND window_handle, unsigned msg, WPARAM wparam, LPARAM lparam) ;
static uint32_t wl_get_token(LPTSTR in, LPTSTR out) ;
static uint32_t str2len(LPTSTR str) ;
LPTSTR GetHelp(uint32_t clpnr, LPTSTR strng, uint32_t len) ;
static void decode_arraystring(LPTSTR strng, INIVALUE *inival) ;
static void generate_arraystring(LPTSTR strng, size_t strng_size, INIVALUE *inival) ;
static void DisplayHelp(bool show_sysparms) ;

#ifndef _CONSOLE
/* These functions make sense only in windowed mode */
/*
+-----------------------------------------------------------------------------------------------------------------------+
|       WinLibMessage                                                                                                   |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| This function processes some of the dialog messages.                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
INT_PTR WinLibMessage(unsigned msg, WPARAM wparam, LPARAM lparam)
{
INT_PTR retcd ;
WPARAM fwSide ;
LPRECT lprc ;
RECT rect, workarea ;
int32_t xPos ;
int32_t yPos ;
int32_t delta_x, delta_y ;
bool movewin ;
static bool snapped_xl = false ;
static bool snapped_xr = false ;
static bool snapped_yt = false ;
static bool snapped_yb = false ;
static int32_t virtual_x, virtual_y ;
int32_t snap_left, snap_right, snap_top, snap_bottom ;
POINT mouse_cursor ;

   retcd = 0 ; /* return 0 if we don't process the message */

   switch(msg)
   {
   case WM_SIZE:
      retcd = 0 ;
      resize_controls(&coordinates[0], coord_size) ;
   break ;

   case WM_SIZING:
      retcd = true ;
      fwSide = wparam;         // edge of window being sized
      lprc = (LPRECT) lparam;  // screen coordinates of drag rectangle
      if ((fwSide == WMSZ_RIGHT) || (fwSide == WMSZ_BOTTOMRIGHT) || (fwSide == WMSZ_TOPRIGHT))
         if ((lprc->right-lprc->left) < min_size_x) lprc->right = lprc->left + min_size_x ;
      if ((fwSide == WMSZ_LEFT) || (fwSide == WMSZ_BOTTOMLEFT) || (fwSide == WMSZ_TOPLEFT))
         if ((lprc->right-lprc->left) < min_size_x) lprc->left = lprc->right - min_size_x ;
      if ((fwSide == WMSZ_BOTTOM) || (fwSide == WMSZ_BOTTOMRIGHT) || (fwSide == WMSZ_BOTTOMLEFT))
         if ((lprc->bottom-lprc->top) < min_size_y) lprc->bottom = lprc->top + min_size_y ;
      if ((fwSide == WMSZ_TOP) || (fwSide == WMSZ_TOPRIGHT) || (fwSide == WMSZ_TOPLEFT))
         if ((lprc->bottom-lprc->top) < min_size_y) lprc->top = lprc->bottom - min_size_y ;
   break ;

//   case WM_WINDOWPOSCHANGING:
//      retcd = true ;
//      ((WINDOWPOS *) lparam)->x += 30 ;
//      ((WINDOWPOS *) lparam)->flags |= SWP_NOMOVE ; /* don't move the window. Done below in WM_MOVE */
//   break ;

   case WM_MOVE:
      retcd = true ;
      movewin = false ;
      delta_x = 0 ;
      delta_y = 0 ;

      GetWindowRect(MainWindow, &rect) ;
      SystemParametersInfo(SPI_GETWORKAREA, 0, &workarea, 0) ;
      snap_left = workarea.left ;
      snap_right = workarea.right - (rect.right - rect.left) ;
      snap_top = workarea.top ;
      snap_bottom = workarea.bottom - (rect.bottom - rect.top) ;

      xPos = rect.left ; /* horizontal position */
      yPos = rect.top ;  /* vertical position */

      if (!snapped_xr)
      {
         if (snapped_xl)
         {/* already snapped left */
            virtual_x = virtual_x + xPos - snap_left ; /* update virtual position */
            movewin = true ;
            delta_x = snap_left - xPos ;
            if (abs(virtual_x - snap_left) >= snap_x)
            { /* leave snap region */
               delta_x = virtual_x - snap_left ;
               snapped_xl = false ;
            }
         }
         else
         {/* not yet snapped */
            if (abs(xPos - snap_left) < snap_x)
            {
               virtual_x = xPos ; /* set virtual position to current position */
               delta_x = snap_left - xPos ;
               movewin = true ;
               snapped_xl = true ;
            }
         }
      }

      if (!snapped_xl)
      {
         if (snapped_xr)
         {/* already snapped right */
            virtual_x = virtual_x + xPos - snap_right ; /* update virtual position */
            movewin = true ;
            delta_x = snap_right - xPos ;
            if (abs(virtual_x - snap_right) >= snap_x)
            { /* leave snap region */
               delta_x = virtual_x - snap_right ;
               snapped_xr = false ;
            }
         }
         else
         {/* not yet snapped */
            if (abs(xPos - snap_right) < snap_x)
            {
               virtual_x = xPos ; /* set virtual position to current position */
               delta_x = snap_right - xPos ;
               movewin = true ;
               snapped_xr = true ;
            }
         }
      }

      if (!snapped_yb)
      {
         if (snapped_yt)
         {/* already snapped on top */
            virtual_y = virtual_y + yPos - snap_top ; /* update virtual position */
            movewin = true ;
            delta_y = snap_top - yPos ;
            if (abs(virtual_y - snap_top) >= snap_y)
            { /* leave snap region */
               delta_y = virtual_y - snap_top ;
               snapped_yt = false ;
            }
         }
         else
         {/* not yet snapped */
            if (abs(yPos - snap_top) < snap_y)
            {
               virtual_y = yPos ; /* set virtual position to current position */
               delta_y = snap_top - yPos ;
               movewin = true ;
               snapped_yt = true ;
            }
         }
      }

      if (!snapped_yt)
      {
         if (snapped_yb)
         {/* already snapped on bottom */
            virtual_y = virtual_y + yPos - snap_bottom ; /* update virtual position */
            movewin = true ;
            delta_y = snap_bottom - yPos ;
            if (abs(virtual_y - snap_bottom) >= snap_y)
            { /* leave snap region */
               delta_y = virtual_y - snap_bottom ;
               snapped_yb = false ;
            }
         }
         else
         {/* not yet snapped */
            if (abs(yPos - snap_bottom) < snap_y)
            {
               virtual_y = yPos ; /* set virtual position to current position */
               delta_y = snap_bottom - yPos ;
               movewin = true ;
               snapped_yb = true ;
            }
         }
      }

      if (movewin)
      {
         GetCursorPos(&mouse_cursor) ;
         SetCursorPos(mouse_cursor.x + delta_x, mouse_cursor.y + delta_y) ;
      }
   break ;

   }
   return(retcd) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       WinLibInit                                                                                                      |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Initialize WinLib. Call this function once at startup.                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void WinLibInit(void)
{
   snap_x = DEFAULT_SNAP_X ;
   snap_y = DEFAULT_SNAP_Y ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       SaveCoordinates                                                                                                 |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Save the coordinates of the dialog and of the controls.                                                               |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void SaveCoordinates(void)
{
RECT rect ;
// save original coordinates of dialog
   GetWindowRect(*coordinates[0].hwnd, &rect) ;
   coordinates[0].top = rect.top ;
   coordinates[0].right = rect.right ;
   coordinates[0].bottom = rect.bottom ;
   coordinates[0].left = rect.left ;
   min_size_x = rect.right - rect.left ; /* the original dialog size is the minimum size */
   min_size_y = rect.bottom - rect.top ; /* the original dialog size is the minimum size */

// save original coordinates of controls
   get_coords(&coordinates[0], coord_size) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       get_coords                                                                                                      |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Get the coordinates of the controls, which are defined in the coordinates structure.                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void get_coords(COORDS *local_coords, size_t size)
{
RECT rect ;
POINT pntlo ;
POINT pntru ;
uint32_t i ;
HWND hwnd ;
   for (i=1; i<size; i++) /* skip first entry; reserved for main dialog */
   {
      hwnd = (local_coords[i].control == 0) ? *local_coords[i].hwnd : GetDlgItem(*local_coords[i].hwnd, local_coords[i].control) ;
      GetClientRect(hwnd, &rect) ;
      if (GetWindowRect(hwnd, &rect) != 0)
      {
         pntlo.x = rect.left ;
         pntlo.y = rect.top ;
         pntru.x = rect.right ;
         pntru.y = rect.bottom ;
         ScreenToClient(*local_coords[i].hwnd, &pntlo) ;
         ScreenToClient(*local_coords[i].hwnd, &pntru) ;
         local_coords[i].left = pntlo.x + local_coords[i].xoffs ;
         local_coords[i].top = pntlo.y + local_coords[i].yoffs ;
         local_coords[i].right = pntru.x + local_coords[i].xoffs ;
         local_coords[i].bottom = pntru.y + local_coords[i].yoffs ;
      }
      else
      {
         local_coords[i].left = 0 ;
         local_coords[i].top = 0 ;
         local_coords[i].right = 0 ;
         local_coords[i].bottom = 0 ;
      }
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       resize_controls                                                                                                 |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| This function is called as a response to the WM_SIZE message. It resizes the controls of a dialog.                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void resize_controls(COORDS *local_coords, size_t size)
{
RECT rect ;
int addx, addy, movx, movy ;
int hsize, vsize, hsize_org, vsize_org ;
int static hsize_prev = 0 ;
int static vsize_prev = 0 ;
size_t i ;
HWND hwnd ;
// get current size of dialog
   GetWindowRect(*local_coords[0].hwnd, &rect) ;
   hsize = rect.right - rect.left ;
   vsize = rect.bottom - rect.top ;
   hsize_org = local_coords[0].right-local_coords[0].left ;
   vsize_org = local_coords[0].bottom-local_coords[0].top ;

   if (hsize_prev == 0) hsize_prev = hsize_org ;
   if (vsize_prev == 0) vsize_prev = vsize_org ;

/* horizontal sizing */
   if (hsize_prev != hsize)
   {
/* Controls must be sorted so that leftmost controls are moved first, when dialog gets smaller and rightmost
   controls are moved first when dialog gets larger */
      for (i=1; i<size; i++) sorted_coord[i-1] = &local_coords[i] ;
      if (hsize < hsize_prev)
      { /* dialog gets smaller */
         qsort(&sorted_coord[0], i-1, sizeof(COORDS *), compare_hup) ;
      }
      else
      { /* dialog gets larger */
         qsort(&sorted_coord[0], i-1, sizeof(COORDS *), compare_hdown) ;
      }

      for (i=1; i<size; i++)
      {
         addx = 0 ;
         addy = 0 ;
         movx = 0 ;
         movy = 0 ;
         if (sorted_coord[i-1]->resizespeed_x != 0) addx = ((hsize - hsize_org) * sorted_coord[i-1]->resizespeed_x) / 1000 ;
         if (sorted_coord[i-1]->movespeed_x != 0) movx = ((hsize-hsize_org) * sorted_coord[i-1]->movespeed_x) / 1000 ;
         if (sorted_coord[i-1]->resizespeed_y != 0) addy = ((vsize - vsize_org) * sorted_coord[i-1]->resizespeed_y) / 1000 ;
         if (sorted_coord[i-1]->movespeed_y != 0) movy = ((vsize-vsize_org) * sorted_coord[i-1]->movespeed_y) / 1000 ;
         hwnd = (sorted_coord[i-1]->control == 0) ? *sorted_coord[i-1]->hwnd : GetDlgItem(*sorted_coord[i-1]->hwnd, sorted_coord[i-1]->control) ;
         MoveWindow(
         hwnd,                                                  // handle of window
         sorted_coord[i-1]->left+movx,                          // horizontal position
         sorted_coord[i-1]->top+movy,                           // vertical position
         sorted_coord[i-1]->right-sorted_coord[i-1]->left+addx, // width
         sorted_coord[i-1]->bottom-sorted_coord[i-1]->top+addy, // height
         true);                                                 // repaint flag
      }
   }
   hsize_prev = hsize ;

/* vertical sizing */
   if (vsize_prev != vsize)
   {
/* Controls must be sorted so that topmost controls are moved first, when dialog gets smaller and bottommost
   controls are moved first when dialog gets larger */
      for (i=1; i<size; i++) sorted_coord[i-1] = &local_coords[i] ;
      if (vsize < vsize_prev)
      { /* dialog gets smaller */
         qsort(&sorted_coord[0], i-1, sizeof(COORDS *), compare_vup) ;
      }
      else
      { /* dialog gets larger */
         qsort(&sorted_coord[0], i-1, sizeof(COORDS *), compare_vdown) ;
      }
      for (i=1; i<size; i++)
      {
         addx = 0 ;
         addy = 0 ;
         movx = 0 ;
         movy = 0 ;
         if (sorted_coord[i-1]->resizespeed_x != 0) addx = ((hsize - hsize_org) * sorted_coord[i-1]->resizespeed_x) / 1000 ;
         if (sorted_coord[i-1]->movespeed_x != 0) movx = ((hsize-hsize_org) * sorted_coord[i-1]->movespeed_x) / 1000 ;
         if (sorted_coord[i-1]->resizespeed_y != 0) addy = ((vsize - vsize_org) * sorted_coord[i-1]->resizespeed_y) / 1000 ;
         if (sorted_coord[i-1]->movespeed_y != 0) movy = ((vsize-vsize_org) * sorted_coord[i-1]->movespeed_y) / 1000 ;
         hwnd = (sorted_coord[i-1]->control == 0) ? *sorted_coord[i-1]->hwnd : GetDlgItem(*sorted_coord[i-1]->hwnd, sorted_coord[i-1]->control) ;
         MoveWindow(
         hwnd,                                                  // handle of window
         sorted_coord[i-1]->left+movx,                          // horizontal position
         sorted_coord[i-1]->top+movy,                           // vertical position
         sorted_coord[i-1]->right-sorted_coord[i-1]->left+addx, // width
         sorted_coord[i-1]->bottom-sorted_coord[i-1]->top+addy, // height
         true);                                                 // repaint flag
      }
   }
   vsize_prev = vsize ;

}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       int compare_hup(const void *op1, const void *op2 )                                                              |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Compare for qsort                                                                                                     |
| For horizontal moves to the left                                                                                      |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
int32_t compare_hup(const void *op1, const void *op2)
{
COORDS *p1, *p2 ;
   p1 = * (COORDS **) op1 ;
   p2 = * (COORDS **) op2 ;
   return (p1->left - p2->left) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       int compare_hdown(const void *op1, const void *op2 )                                                            |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Compare for qsort                                                                                                     |
| For horizontal moves to the right                                                                                     |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
int32_t compare_hdown(const void *op1, const void *op2)
{
COORDS *p1, *p2 ;
   p1 = * (COORDS **) op1 ;
   p2 = * (COORDS **) op2 ;
   return (p2->left - p1->left) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       int compare_vup(const void *op1, const void *op2 )                                                              |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Compare for qsort                                                                                                     |
| For vertical moves to the top                                                                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
int32_t compare_vup(const void *op1, const void *op2)
{
COORDS *p1, *p2 ;
   p1 = * (COORDS **) op1 ;
   p2 = * (COORDS **) op2 ;
   return (p1->top - p2->top) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       int compare_vdown(const void *op1, const void *op2 )                                                            |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Compare for qsort                                                                                                     |
| For vertical moves to the bottom                                                                                      |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
int32_t compare_vdown(const void *op1, const void *op2)
{
COORDS *p1, *p2 ;
   p1 = * (COORDS **) op1 ;
   p2 = * (COORDS **) op2 ;
   return (p2->top - p1->top) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       DlgHelp                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Help Dialog                                                                                                           |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
INT_PTR CALLBACK DlgHelp(HWND window_handle, unsigned msg, WPARAM wparam, LPARAM lparam)
{
uint32_t retcd ; /* return nonzero, if the message is processed here and zero, if it is not */

   lparam = lparam ; /* to avoid warning with warning level 4 */
   retcd = 0 ; /* return 0 if we don't process the message */
   switch(msg)
   {
   case WM_INITDIALOG:
      retcd = 1 ;
      HelpWindow = window_handle ;
      DisplayHelp(false) ;
   break ;

   case WM_COMMAND:
      retcd = 1 ;
      switch(LOWORD(wparam))
      {
      case WM_SYSPARM:
         DisplayHelp(true) ;
      break ;

      case IDOK:
      case IDABORT:
      case IDCANCEL:
         EndDialog(HelpWindow, true);      /* exits the dialog box */
      break ;

      default:
         retcd = 0 ; /* we have not processed this command, so return zero */
      break ;
      }
   break ;
   }

   return(retcd) ;
} /* DlgHelp */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       DisplayHelp(bool show_sysparms)                                                                                 |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Display Help in open window.                                                                                          |
| System command line parameters will be displayed if sysparms is true.                                                 |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void DisplayHelp(bool show_sysparms)
{
HWND msg_hwnd ;
uint32_t i ;
bool parms_supported ;
   parms_supported = false ;
   msg_hwnd = GetDlgItem(HelpWindow, helpdlgitem) ;
   if (HelpWindowName != NULL) SendMessage(HelpWindow, WM_SETTEXT, 0, (LPARAM) HelpWindowName);
   SendMessage(msg_hwnd, EM_SETSEL, 0, -1); /* clear window */
   SendMessage(msg_hwnd, EM_REPLACESEL, (BOOL) false, (WPARAM) help_header) ;
   SendMessage(msg_hwnd, EM_REPLACESEL, (BOOL) false, (WPARAM) purpose) ;
   SendMessage(msg_hwnd, EM_REPLACESEL, (BOOL) false, (WPARAM) _T("The following command line parameters are supported:") NEWLINE) ;
   GetHelp_pos = 1 ;
   for (i=0; i<clo_size; i++)
   {
      if ((show_sysparms) || ((clo[i].parmname == NULL) || (clo[i].parmname[0] != SYSP[0])))
      {
         SendMessage(msg_hwnd, EM_REPLACESEL, (BOOL) false, (WPARAM) GetHelp(i, local_strng1, STRINGSIZE)) ;
         parms_supported = true ;
      }
   }
   if (!parms_supported) SendMessage(msg_hwnd, EM_REPLACESEL, (BOOL) false, (WPARAM) _T("=== none ===") NEWLINE) ;
}
#endif /* _CONSOLE */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       GetHelp                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Get Help for a specific command line parameter                                                                        |
| len is the maximum number of characters in strng                                                                      |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
LPTSTR GetHelp(uint32_t clpnr, LPTSTR strng, uint32_t len)
{
LPTSTR prefix ;
LPTSTR parmname ;
LPTSTR parmval ;
TCHAR parm[16] ;

   strng[0] = _T('\0') ;
   if (clo[clpnr].help != NULL)
   {
      switch (clo[clpnr].parmtype)
      {
         case PARM_SIMPLE: prefix = _T("-") ; parmname = clo[clpnr].parmname ; parmval = _T("") ; break ;
         case PARM_INT: prefix = _T("-") ; parmname = clo[clpnr].parmname ; parmval = _T("=n") ; break ;
         case PARM_STRING: prefix = _T("-") ; parmname = clo[clpnr].parmname ; parmval = _T("=\"string\"") ; break ;
         case PARM_POS: prefix = _T("") ; _stprintf_s(parm, sizeof(parm)/sizeof(parm[0]), _T("pos%u"), GetHelp_pos) ; parmname = parm ; GetHelp_pos++; parmval = _T("") ; break ;
         default: prefix = _T("") ; parmname = _T("") ; parmval = _T("") ; break ;
      }
      _stprintf_s(strng, len, _T("%s%s%s: %s") NEWLINE, prefix, parmname, parmval, clo[clpnr].help) ;
   }

   return(strng) ;
} /* GetHelp */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       PrintProgramInfo                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Open a message box and print some information about the program                                                       |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void PrintProgramInfo(uint32_t dlgitem, LPTSTR local_pgmname, LPTSTR date, LPTSTR time)
{
#ifdef _CONSOLE
uint32_t i ;
#endif
   helpdlgitem = dlgitem ;
   _stprintf_s(help_header, STRINGSIZE, _T("%s") NEWLINE
                 _T("Written by: ") PROGRAMMER _T(", compiled on %s, %s") NEWLINE
                 _T("for updates check: ") WEBSITE NEWLINE
                 _T("or email to: ") EMAILADDR NEWLINE NEWLINE,
                 local_pgmname, date, time) ;
   _stprintf_s(HelpWindowName, STRINGSIZE, _T("%s - Program Information"), local_pgmname) ;
#ifdef _CONSOLE
   printf(_T("%s"), help_header) ;
   printf(_T("Purpose: %s\n"), purpose) ;
   printf(_T("The following command line parameters are supported:\n")) ;
   for (i=0; i<clo_size; i++) printf("%s", GetHelp(i, local_strng1, STRINGSIZE)) ;
#else
   DialogBox(MyInstance, _T("DLG_HELP"), MainWindow, (DLGPROC)DlgHelp) ;
#endif
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       PrintInfo                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Open a message box and print info.                                                                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#ifndef _CONSOLE
void PrintInfo(LPTSTR info)
{
   info = info ; /* to avoid warning with warning level 4 */
   DialogBox(MyInstance, _T("DLG_INFO"), MainWindow, (DLGPROC)DlgInfo) ;
}
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       my_malloc                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Private malloc procedure. Print error message including file name and line number.                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void *my_malloc(size_t size, LPTSTR file, int line, int aux)
{
void *ptr ;
int i ;
   ptr = malloc(size) ;
   if (ptr == NULL)
   {
      i = (int) _tcsclen(file) ;
      while ((i > 0) && (file[i-1] != _T('\\'))) i-- ;
      _stprintf_s(Info_Strng, STRINGSIZE,
             _T("%s\r\r\n")
             _T("Version %u.%2.2u\r\r\n\r\r\n")
             _T("Heap overflow in \"%s\" at line %i, aux=%i\r\r\n")
             _T("malloc has returned a NULL pointer.\r\r\n\r\r\n")
             _T("This program will now terminate."), programname, programversion/100, programversion%100, &file[i], line, aux) ;
#ifdef _CONSOLE
      printf("%s", Info_Strng) ;
#else
      PrintInfo(Info_Strng) ;
#endif
      exit(ERROR) ;
   }
   return (ptr) ;
}


/*
+-----------------------------------------------------------------------------------------------------------------------+
|       my_free                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Private free procedure.                                                                                               |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void my_free(void *ptr)
{
   if (ptr != NULL) free(ptr) ;
   return ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       DlgInfo                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Info Dialog                                                                                                           |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
INT_PTR CALLBACK DlgInfo(HWND window_handle, unsigned msg, WPARAM wparam, LPARAM lparam)
{
INT_PTR retcd ; /* return true, if the message is processed here and false, if it is not (except WM_INITDIALOG) */

   lparam = lparam ; /* to avoid warning with warning level 4 */
   retcd = 0 ; /* return 0 if we don't process the message */
   switch(msg)
   {
   case WM_INITDIALOG:
      retcd = 1 ;
      SendMessage(window_handle, WM_SETTEXT, 0, (LPARAM) _T("Info"));
      SendMessage(GetDlgItem(window_handle, DLG_INFOLIST), WM_SETTEXT, 0, (WPARAM) Info_Strng) ;
      retcd = FALSE ;
   break ;

   case WM_COMMAND:
      retcd = 1 ;
      switch(LOWORD(wparam))
      {
      case IDOK:
      case IDABORT:
      case IDCANCEL:
         EndDialog(window_handle, true);      /* exits the dialog box */
      break ;

      default:
         retcd = 0 ; /* we have not processed this command, so return zero */
      break ;
      }
   break ;
   }

   return(retcd) ;
} /* DlgInfo */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       wl_get_token                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Get the next token from the input string                                                                              |
|                                                                                                                       |
| Returns: Number of bytes taken from input string.                                                                     |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint32_t wl_get_token(LPTSTR in, LPTSTR out)
{
uint32_t i, j ;
bool terminate, hyphen ;
   i = 0 ;
   j = 0 ;
   terminate = false ;
   hyphen = false ;
   while (!terminate)
   {
      out[j] = _T('\0') ;
      if (in[i] == _T('\0')) terminate = true ;
      if ((in[i] == _T(' ')) && !hyphen)
      {
         out[j] = _T('\0') ;
         i++ ;
         terminate = true ;
      }
      if ((in[i] == _T('"')) && hyphen)
      {
         out[j] = _T('\0') ;
         i++ ;
         terminate = true ;
      }
      if (!terminate)
      {
         out[j] = in[i] ;
         hyphen = hyphen || (in[i] == _T('"')) ;
         if (in[i] != _T('"')) j++ ;
         i++ ;
      }
   }
   return(i) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       getparms                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Get the command line parameters                                                                                       |
|                                                                                                                       |
| Returns pointer to verbal error description or NULL if no error found.                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
LPTSTR getparms(LPTSTR cmdline)
{
bool cmd_done, found_eq ;
static TCHAR token[MAXPARMLENGTH] ;
LPTSTR parmvalue ;
uint32_t j, k ;
size_t offset ;
int32_t intval ;
bool found ;
uint32_t increment ;
   cmd_done = false ;
   errorparm[0] = _T('\0') ;
   parmvalue = NULL ;
   increment = 0 ;
/* check parameters if cmdline points to a string */
   if ((cmdline != NULL) && (cmdline[0] != _T('\0')))
   {
      offset = 0 ;
      while (!cmd_done)
      {
         while (cmdline[offset] == _T(' ')) offset++ ; /* skip any possibly leading blanks */
         if (cmdline[offset] == _T('\0')) cmd_done = true ;
         else
         {
            if ((cmdline[offset] == _T('-')) || (cmdline[offset] == _T('/')))
            {/* this is a parameter */
               offset++ ;
               if ((cmdline[offset] == _T(' ')) || (cmdline[offset] == _T('\0'))) /* we found a "-" without a parameter */
               {
                  token[0] = _T('\0') ;
               }
               else
               {
                  if ((increment = wl_get_token(&cmdline[offset], &token[0])) == 0) cmd_done = true ;
               }
               if (!cmd_done)
               { /* we isolated a parameter. Check if it is an assignment. */
                  found_eq = false ;
                  k = 0 ;
                  while ((token[k] != _T('\0')) && (token[k] != _T('='))) k++ ;
                  if (token[k] == _T('='))
                  {
                     found_eq = true ;
                     token[k] = _T('\0') ; /* terminator for parameter name */
                     parmvalue = &token[k+1] ;
                  }
                  /* Now we search the parameter name in the list. */
                  found = false ;
                  for (j=0; (j<clo_size) && (!found); j++)
                  {
                     if ((clo[j].parmtype == PARM_SIMPLE)
                      || (clo[j].parmtype == PARM_INT)
                      || (clo[j].parmtype == PARM_STRING))
                     found = lstrcmp(token, clo[j].parmname) == 0 ;
                  }
                  if (found_eq) token[k] = _T('=') ; /* restore = */
                  if (found)
                  {
                     j-- ; /* fix increment from previous for loop */
                     if (clo[j].parmtype == PARM_SIMPLE) clo[j].found = true ;
                     if (clo[j].parmtype == PARM_INT)
                     { /* convert parmvalue to int */
                        if (clo[j].found == true)
                        {
                           _stprintf_s(&errorparm[_tcsclen(errorparm)], MAXPARMLENGTH-((size_t) _tcsclen(errorparm)), _T("%c%s (duplicate parameter) "), cmdline[offset-1], token) ;
                           clo[j].found = false ;
                        }
                        if (_stscanf_s(parmvalue, _T("%i"), &intval) == 1)
                        {
                           clo[j].value = intval ;
                           clo[j].found = true ;
                        }
                        else
                        { /* we support some special values: off=0, on=1 */
                           if (_tcsicmp(parmvalue, _T("off")) == 0)
                           {
                              clo[j].value = 0 ;
                              clo[j].found = true ;
                           }
                           if (_tcsicmp(parmvalue, _T("on")) == 0)
                           {
                              clo[j].value = 1 ;
                              clo[j].found = true ;
                           }
                        }
                        if (!(clo[j].found))
                           _stprintf_s(&errorparm[_tcsclen(errorparm)], MAXPARMLENGTH-((size_t) _tcsclen(errorparm)), _T("%c%s (invalid value)"), cmdline[offset-1], token) ;
                     }
                     if (clo[j].parmtype == PARM_STRING)
                     { /* copy parmvalue to destination */
                        if (clo[j].value != 0)
                        {
                           _tcsncpy_s(clo[j].dest, clo[j].value, parmvalue, clo[j].value) ;
                           ((LPTSTR) clo[j].dest)[clo[j].value-1] = _T('\0') ; /* force string termination */
                           clo[j].found = true ;
                        }
                     }
                  }
                  else
                  { /* unexpected parameter */
                     _stprintf_s(&errorparm[_tcsclen(errorparm)], MAXPARMLENGTH-((size_t) _tcsclen(errorparm)), _T("%c%s "), cmdline[offset-1], token) ;
                  }
                  offset += increment ;
               }
            }
            else
            { /* this is a positional parameter (filename e.t.c.) */
               if ((increment = wl_get_token(&cmdline[offset], &token[0])) == 0) cmd_done = true ;
               /* Now we search the next unassigned positional parameter in the list. */
               found = false ;
               for (j=0; (j<clo_size) && (!found); j++)
               {
                  if ((clo[j].parmtype == PARM_POS) && (clo[j].found == false) && (clo[j].dest != NULL))
                  {
                     _tcsncpy_s(clo[j].dest, clo[j].value, token, clo[j].value) ;
                     clo[j].found = true ;
                     found = true ;
                  }
               }
               offset += increment ;
            }
         }
      }
   }
   offset = _tcsclen(errorparm) ;
   if (offset > 0) offset-- ;
   if (errorparm[offset] == _T(' ')) errorparm[offset] = _T('\0') ; /* remove trailing blank for the beauty of the error message */
   return(offset==0 ? NULL : errorparm) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       get_registry                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Get the required data from registry. Return true if all keys were successfully read.                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool get_registry(MYINI *myini)
{
HKEY keyhandle ;
LONG retv ;
bool rc ;
DWORD csize ;
uint32_t i, is, id ;
int32_t sd1, sd2 ;
TCHAR skn[SKN_LENGTH] ; /* sub key name */
LPTSTR gr_buffer ;
LPTSTR destination ;
bool terminate ;

   rc = false ;
   gr_buffer = (LPTSTR) malloc(DYN_BUFFERCHARSIZE * sizeof(TCHAR)) ;
   if (gr_buffer != NULL)
   { /* Execute function only if malloc succeeds. Otherwise something is seriously wrong */
      if ((myini->initype & INI_REGREAD) != 0)
      {
         rc = true ;
         for (i=0; i<myini->size ; i++)
         {
// Open Registry key
// assemble sub key name
            _tcscpy_s(skn, SKN_LENGTH, myini->inikeyname) ;
            if ((myini->inivalues[i].sectionname != NULL) && (_tcsclen(myini->inivalues[i].sectionname) != 0))
            {
               _tcscat_s(skn, SKN_LENGTH, _T("\\")) ;
               _tcscat_s(skn, SKN_LENGTH, myini->inivalues[i].sectionname) ;
            }
            retv = RegOpenKeyEx(myini->hkey,
                   skn,
                   0,
                   KEY_READ,
                   &keyhandle) ;
            if (retv != ERROR_SUCCESS) rc = false ;
            else
            {
               switch (myini->inivalues[i].valuetype)
               {
                  case REG_DWORD :
                  if (((myini->inivalues[i].flags) & REGFLAG_ARRAY) == 0)
                  { /* This is a single DWORD */
                     csize = 4 ;
                     retv = RegQueryValueEx(
                            keyhandle,                         // handle of key to query
                            myini->inivalues[i].valuename,     // address of name of value to query
                            NULL,                              // reserved
                            NULL,                              // address of buffer for value type
                            (LPBYTE)myini->inivalues[i].value, // address of data buffer
                            &csize                             // address of data buffer size
                            ) ;
                     if (retv != ERROR_SUCCESS) rc = false ;
                  }
                  else
                  { /* This is an array of DWORDs stored as a character string */
                     csize = DYN_BUFFERCHARSIZE * sizeof(TCHAR) ;
                     retv = RegQueryValueEx(
                            keyhandle,                         // handle of key to query
                            myini->inivalues[i].valuename,     // address of name of value to query
                            NULL,                              // reserved
                            NULL,                              // address of buffer for value type
                            (LPBYTE) gr_buffer,                // address of data buffer
                            &csize                             // address of data buffer size in number of bytes
                            ) ;
                     if (retv != ERROR_SUCCESS) rc = false ;
                     else
                     {
                        decode_arraystring(gr_buffer, &myini->inivalues[i]) ;
                     }
                  }
                  break ;

                  case REG_QWORD :
                  if (((myini->inivalues[i].flags) & REGFLAG_ARRAY) == 0)
                  { /* This is a single QWORD */
                     csize = 8 ;
                     retv = RegQueryValueEx(
                            keyhandle,                         // handle of key to query
                            myini->inivalues[i].valuename,     // address of name of value to query
                            NULL,                              // reserved
                            NULL,                              // address of buffer for value type
                            (LPBYTE)myini->inivalues[i].value, // address of data buffer
                            &csize                             // address of data buffer size
                            ) ;
                     if (retv != ERROR_SUCCESS) rc = false ;
                  }
                  else
                  { /* This is an array of QWORDs stored as a character string */
                     csize = DYN_BUFFERCHARSIZE * sizeof(TCHAR) ;
                     retv = RegQueryValueEx(
                            keyhandle,                         // handle of key to query
                            myini->inivalues[i].valuename,     // address of name of value to query
                            NULL,                              // reserved
                            NULL,                              // address of buffer for value type
                            (LPBYTE) gr_buffer,                // address of data buffer
                            &csize                             // address of data buffer size in number of bytes
                            ) ;
                     if (retv != ERROR_SUCCESS) rc = false ;
                     else
                     {
                        decode_arraystring(gr_buffer, &myini->inivalues[i]) ;
                     }
                  }
                  break ;

                  case REG_SZ :
                  case REG_MULTI_SZ :
                  csize = (DWORD) myini->inivalues[i].valuesize * sizeof(TCHAR) ;
                  retv = RegQueryValueEx(
                         keyhandle,                         // handle of key to query
                         myini->inivalues[i].valuename,     // address of name of value to query
                         NULL,                              // reserved
                         NULL,                              // address of buffer for value type
                         (LPBYTE)myini->inivalues[i].value, // address of data buffer
                         &csize                             // address of data buffer size in number of bytes
                         ) ;
                  if (retv != ERROR_SUCCESS) rc = false ;
                  break ;
               }

               RegCloseKey(keyhandle) ;
            }
         }
      }

      if ((myini->initype & INI_FILEREAD) != 0)
      {
         rc = true ;
         _tcscpy_s(inifilename, MAX_PATH+1, myini->inifilename) ;
         for (i=0; i<myini->size ; i++)
         {
            switch (myini->inivalues[i].valuetype)
            {
               case REG_DWORD :
               /* This may be a single DWORD or an array of DWORDs stored as a character string */
                  if (((myini->inivalues[i].flags) & REGFLAG_ARRAY) == 0)
                  { /* This is a single DWORD. Earlier versions did not include the size, so we set it to 1 */
                     myini->inivalues[i].valuesize = 1 ;
                  }
                  csize = DYN_BUFFERCHARSIZE ; /* number of characters, not bytes! */
                  sd1 = GetPrivateProfileString(
                         myini->inivalues[i].sectionname,  // pointer to section name
                         myini->inivalues[i].valuename,    // pointer to key name
                         _T("~~"),                         // points to default string
                         gr_buffer,                        // points to destination buffer
                         csize,                            // size of destination buffer in number of characters
                         inifilename                       // pointer to initialization filename
                         ) ;
                  csize = DYN_BUFFERCHARSIZE ; /* number of characters, not bytes! */
                  sd2 = GetPrivateProfileString(
                         myini->inivalues[i].sectionname,  // pointer to section name
                         myini->inivalues[i].valuename,    // pointer to key name
                         _T("####"),                       // points to default string
                         gr_buffer,                        // points to destination buffer
                         csize,                            // size of destination buffer in number of characters
                         inifilename                       // pointer to initialization filename
                         ) ;
                  if (sd1 == sd2)
                  { /* string was found */
                     decode_arraystring(gr_buffer, &myini->inivalues[i]) ;
                  }
                  else
                     rc = false ; /* key was not found */
               break ;

               case REG_QWORD :
               /* This may be a single QWORD or an array of QWORDs stored as a character string */
                  if (((myini->inivalues[i].flags) & REGFLAG_ARRAY) == 0)
                  { /* This is a single DWORD. Earlier versions did not include the size, so we set it to 1 */
                     myini->inivalues[i].valuesize = 1 ;
                  }
                  csize = DYN_BUFFERCHARSIZE ; /* number of characters, not bytes! */
                  sd1 = GetPrivateProfileString(
                         myini->inivalues[i].sectionname,  // pointer to section name
                         myini->inivalues[i].valuename,    // pointer to key name
                         _T("~~"),                         // points to default string
                         gr_buffer,                        // points to destination buffer
                         csize,                            // size of destination buffer in number of characters
                         inifilename                       // pointer to initialization filename
                         ) ;
                  csize = DYN_BUFFERCHARSIZE ; /* number of characters, not bytes! */
                  sd2 = GetPrivateProfileString(
                         myini->inivalues[i].sectionname,  // pointer to section name
                         myini->inivalues[i].valuename,    // pointer to key name
                         _T("####"),                       // points to default string
                         gr_buffer,                        // points to destination buffer
                         csize,                            // size of destination buffer in number of characters
                         inifilename                       // pointer to initialization filename
                         ) ;
                  if (sd1 == sd2)
                  { /* string was found */
                     decode_arraystring(gr_buffer, &myini->inivalues[i]) ;
                  }
                  else
                     rc = false ; /* key was not found */
               break ;

               case REG_SZ :
                  csize = DYN_BUFFERCHARSIZE ; /* number of characters, not bytes! */
                  sd1 = GetPrivateProfileString(
                         myini->inivalues[i].sectionname,  // pointer to section name
                         myini->inivalues[i].valuename,    // pointer to key name
                         _T("~~"),                         // points to default string
                         gr_buffer,                        // points to destination buffer
                         csize,                            // size of destination buffer in number of characters
                         inifilename                       // pointer to initialization filename
                         ) ;
                  csize = DYN_BUFFERCHARSIZE ;
                  sd2 = GetPrivateProfileString(
                         myini->inivalues[i].sectionname,  // pointer to section name
                         myini->inivalues[i].valuename,    // pointer to key name
                         _T("####"),                       // points to default string
                         gr_buffer,                        // points to destination buffer
                         csize,                            // size of destination buffer in number of characters
                         inifilename                       // pointer to initialization filename
                         ) ;
                  csize = (DWORD) myini->inivalues[i].valuesize ;
                  if (sd1 == sd2)
                  { /* string was found */
                     if (csize > DYN_BUFFERCHARSIZE) csize = DYN_BUFFERCHARSIZE ; /* truncate string if too long */
                     _tcsncpy_s(myini->inivalues[i].value, myini->inivalues[i].valuesize, gr_buffer, csize) ; /* key was found */
                     *(((LPTSTR) myini->inivalues[i].value) + csize - 1) = _T('\0') ;
                  }
                  else
                     rc = false ; /* key was not found */
               break ;

               case REG_MULTI_SZ :
                  sd1 = GetPrivateProfileString(
                         myini->inivalues[i].sectionname,  // pointer to section name
                         myini->inivalues[i].valuename,    // pointer to key name
                         _T("~~"),                         // points to default string
                         gr_buffer,                        // points to destination buffer
                         DYN_BUFFERCHARSIZE,               // size of destination buffer in number of characters
                         inifilename                       // pointer to initialization filename
                         ) ;
                  sd2 = GetPrivateProfileString(
                         myini->inivalues[i].sectionname,  // pointer to section name
                         myini->inivalues[i].valuename,    // pointer to key name
                         _T("####"),                       // points to default string
                         gr_buffer,                        // points to destination buffer
                         DYN_BUFFERCHARSIZE,               // size of destination buffer in number of characters
                         inifilename                       // pointer to initialization filename
                         ) ;
                  if (sd1 == sd2)
                  { /* string was found. */
                     destination = myini->inivalues[i].value ;
//                     _tcscpy_s(gr_buffer, sizeof(gr_buffer), destination) ; /* get the original data to buffer */
                     is = 0 ;
                     id = 0 ;
                     terminate = false ;
                     do
                     {
                        if (gr_buffer[is] == _T('\0'))
                        {
                           destination[id+0] = _T('\0') ;
                           destination[id+1] = _T('\0') ;
                           terminate = true ;
                        }
                        else
                        {
                           if (gr_buffer[is] == ESCAPE_CHAR)
                           {
                              if (gr_buffer[is+1] == ESCAPE_CHAR)
                              {
                                 destination[id] = ESCAPE_CHAR ;
                              }
                              if (gr_buffer[is+1] == MULTI_SZ_CHAR)
                              {
                                 destination[id] = _T('\0') ;
                              }
                              id++ ;
                              is += 2 ;
                           }
                           else
                           { /* copy character */
                              destination[id] = gr_buffer[is] ;
                              id++ ;
                              is++ ;
                           }
                        }
                     } while (!terminate) ;
                  }
                  else
                     rc = false ; /* key was not found */
               break ;
            }
         }
      }
      free(gr_buffer) ;
   }
   return(rc) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       put_registry                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Put data into registry                                                                                                |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void put_registry(MYINI *myini)
{
HKEY keyhandle ;
LONG retv ;
DWORD csize ;
uint32_t disposition ;
uint32_t i, is, id ;
bool value_exists, terminate, rc ;
TCHAR skn[SKN_LENGTH] ; /* sub key name */
LPTSTR source ;
LPTSTR pr_buffer ;
int32_t sd1, sd2 ;

   pr_buffer = (LPTSTR) malloc(DYN_BUFFERCHARSIZE * sizeof(TCHAR)) ;
   if (pr_buffer != NULL)
   { /* Execute function only if malloc succeeds. Otherwise something is seriously wrong */
      if ((myini->initype & INI_REGWRITE) != 0)
      {
         for (i=0; i<myini->size; i++)
         {
//    assemble sub key name
            _tcscpy_s(skn, _countof(skn), myini->inikeyname) ;
            if ((myini->inivalues[i].sectionname != NULL) && (_tcsclen(myini->inivalues[i].sectionname) != 0))
            {
               _tcscat_s(skn, SKN_LENGTH, _T("\\")) ;
               _tcscat_s(skn, SKN_LENGTH, myini->inivalues[i].sectionname) ;
            }
//    Open or Create the key
            retv = RegCreateKeyEx(myini->hkey,   // handle of an open key
                   skn,                          // address of subkey name
                   0,                            // reserved
                   NULL,                         // address of class string
                   REG_OPTION_NON_VOLATILE,      // special options flag
                   KEY_ALL_ACCESS,               // desired security access
                   REG_OPTION_NON_VOLATILE,      // address of key security structure
                   &keyhandle,                   // address of buffer for opened handle
                   (LPDWORD) &disposition        // address of disposition value buffer
               ) ;
            if (retv == ERROR_SUCCESS)
            { // write data unless read only value and not newly created
               value_exists = (RegQueryValueEx(keyhandle, myini->inivalues[i].valuename, NULL, NULL, NULL, NULL) == ERROR_SUCCESS) ;
               if ((((myini->inivalues[i].flags) & REGFLAG_WRITEBACK) != 0) || (!value_exists))
               {
                  switch (myini->inivalues[i].valuetype)
                  {
                     case REG_DWORD :
                        if (((myini->inivalues[i].flags) & REGFLAG_ARRAY) == 0)
                        { /* This is a normal DWORD */
                           retv = RegSetValueEx(
                                  keyhandle,                               // handle of key to set value for
                                  myini->inivalues[i].valuename,           // address of name of value to set
                                  0,                                       // reserved
                                  REG_DWORD,                               // flag for value type
                                  (CONST BYTE *)myini->inivalues[i].value, // address of data
                                  4                                        // size of value data in number of bytes
                                  ) ;
                        }
                        else
                        { /* This is an array of DWORDs to be stored as a character string */
                           generate_arraystring(pr_buffer, DYN_BUFFERCHARSIZE, &myini->inivalues[i]) ;
                           csize = ((DWORD) _tcsclen(pr_buffer) + 1) * sizeof(TCHAR) ;
                           retv = RegSetValueEx(
                                  keyhandle,                        // handle of key to set value for
                                  myini->inivalues[i].valuename,    // address of name of value to set
                                  0,                                // reserved
                                  REG_SZ,                           // flag for value type
                                  (const BYTE *) pr_buffer,         // address of value data
                                  csize                             // size of value data in number of bytes including termination 0
                                  ) ;
                        }
                     break ;

                     case REG_QWORD :
                        if (((myini->inivalues[i].flags) & REGFLAG_ARRAY) == 0)
                        { /* This is a normal QWORD */
                           retv = RegSetValueEx(
                                  keyhandle,                               // handle of key to set value for
                                  myini->inivalues[i].valuename,           // address of name of value to set
                                  0,                                       // reserved
                                  REG_QWORD,                               // flag for value type
                                  (CONST BYTE *)myini->inivalues[i].value, // address of data
                                  8                                        // size of value data in number of bytes
                                  ) ;
                        }
                        else
                        { /* This is an array of QWORDs to be stored as a character string */
                           generate_arraystring(pr_buffer, DYN_BUFFERCHARSIZE, &myini->inivalues[i]) ;
                           csize = ((DWORD) _tcsclen(pr_buffer) + 1) * sizeof(TCHAR) ;
                           retv = RegSetValueEx(
                                  keyhandle,                        // handle of key to set value for
                                  myini->inivalues[i].valuename,    // address of name of value to set
                                  0,                                // reserved
                                  REG_SZ,                           // flag for value type
                                  (const BYTE *) pr_buffer,         // address of value data
                                  csize                             // size of value data in number of bytes including termination 0
                                  ) ;
                        }
                     break ;

                     case REG_SZ :
                        csize = ((DWORD) _tcsclen(myini->inivalues[i].value) + 1) * sizeof(TCHAR) ;
                        retv = RegSetValueEx(
                               keyhandle,                        // handle of key to set value for
                               myini->inivalues[i].valuename,    // address of name of value to set
                               0,                                // reserved
                               REG_SZ,                           // flag for value type
                               myini->inivalues[i].value,        // address of value data
                               csize                             // size of value data in number of bytes including termination 0
                               ) ;
                     break ;

                     case REG_MULTI_SZ :
                        csize = (str2len(myini->inivalues[i].value) + 1) * sizeof(TCHAR) ; /* to account for 2nd terminating null character */
                        retv = RegSetValueEx(
                               keyhandle,                        // handle of key to set value for
                               myini->inivalues[i].valuename,    // address of name of value to set
                               0,                                // reserved
                               REG_MULTI_SZ,                     // flag for value type
                               myini->inivalues[i].value,        // address of value data
                               csize                             // size of value data in number of bytes including termination 0
                               ) ;
                     break ;
                  }
               }
               RegCloseKey(keyhandle) ;
            }
         }
      }

      if ((myini->initype & INI_FILEWRITE) != 0)
      {
         _tcscpy_s(inifilename, MAX_PATH+1, myini->inifilename) ;
         for (i=0; i<myini->size ; i++)
         { // write data unless read only value and not newly created
            csize = DYN_BUFFERCHARSIZE ;
            sd1 = GetPrivateProfileString(
                   myini->inivalues[i].sectionname,  // pointer to section name
                   myini->inivalues[i].valuename,    // pointer to key name
                   _T("~~"),                         // points to default string
                   pr_buffer,                        // points to destination buffer
                   csize,                            // size of destination buffer in number of characters
                   inifilename                       // pointer to initialization filename
                   ) ;
            csize = DYN_BUFFERCHARSIZE ;
            sd2 = GetPrivateProfileString(
                   myini->inivalues[i].sectionname,  // pointer to section name
                   myini->inivalues[i].valuename,    // pointer to key name
                   _T("####"),                       // points to default string
                   pr_buffer,                        // points to destination buffer
                   csize,                            // size of destination buffer in number of characters
                   inifilename                       // pointer to initialization filename
                   ) ;
            value_exists = (sd1 == sd2) ; /* string was found */
            if ((((myini->inivalues[i].flags) & REGFLAG_WRITEBACK) != 0) || (!value_exists))
            {
               switch (myini->inivalues[i].valuetype)
               {
                  case REG_DWORD :
                  case REG_QWORD :
                     if (((myini->inivalues[i].flags) & REGFLAG_ARRAY) == 0)
                     { /* This is a single DWORD. Earlier versions did not include the size, so we set it to 1 for upward compatibility reasons. */
                        myini->inivalues[i].valuesize = 1 ;
                     }
                     generate_arraystring(pr_buffer, DYN_BUFFERCHARSIZE, &myini->inivalues[i]) ;
                     rc = WritePrivateProfileString(
                            myini->inivalues[i].sectionname,  // pointer to section name
                            myini->inivalues[i].valuename,    // pointer to key name
                            pr_buffer,                        // pointer to string to add
                            inifilename                       // pointer to initialization filename
                            ) ;
                  break ;

                  case REG_SZ :
                     rc = WritePrivateProfileString(
                            myini->inivalues[i].sectionname,  // pointer to section name
                            myini->inivalues[i].valuename,    // pointer to key name
                            myini->inivalues[i].value,        // pointer to string to add
                            inifilename                       // pointer to initialization filename
                            ) ;
                  break ;

                  case REG_MULTI_SZ :
                    is = 0 ;
                    id = 0 ;
                    source = myini->inivalues[i].value ;
                    terminate = false ;
                    do
                    {
                       if (source[is] != _T('\0'))
                       {
                          if  (source[is] == ESCAPE_CHAR)
                          { /* double the escape character */
                             pr_buffer[id] = ESCAPE_CHAR ;
                             id++ ;
                          }
                          pr_buffer[id] = source[is] ;
                          id++ ;
                          is++ ;
                       }
                       else
                       {
                          if (source[is+1] != _T('\0'))
                          {
                             pr_buffer[id+0] = ESCAPE_CHAR ;
                             pr_buffer[id+1] = MULTI_SZ_CHAR ;
                             id += 2 ;
                             is++ ;
                          }
                          else
                          {
                             terminate = true ;
                             pr_buffer[id] = _T('\0') ;
                          }
                       }
                    } while (!terminate) ;
                    rc = WritePrivateProfileString(
                           myini->inivalues[i].sectionname,  // pointer to section name
                           myini->inivalues[i].valuename,    // pointer to key name
                           pr_buffer,                        // pointer to string to add
                           inifilename                       // pointer to initialization filename
                           ) ;
                  break ;
               }
            }
         }
      }
      free(pr_buffer) ;
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       decode_arraystring                                                                                              |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Convert a character string with multiple WORDs into an array of WORDs                                                 |
| String may contain hex words, which must be preceeded by a "0x".                                                      |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void decode_arraystring(LPTSTR strng, INIVALUE *inival)
{
uint32_t i ;
LPTSTR da_buff ;
LPTSTR next_token = NULL ;
int32_t *val32 ;
int64_t *val64 ;
   val32 =             inival->value ;
   val64 = (int64_t *) inival->value ;
   da_buff = _tcstok_s(strng, _T(",.; "), &next_token) ;
   for (i=0; i<inival->valuesize; i++)
   {
      if (da_buff != NULL)
      {
         if ((da_buff[0] == _T('0')) && ((da_buff[1] == _T('x')) || (da_buff[1] == _T('X'))))
         {
            if (inival->valuetype == REG_DWORD) _stscanf_s(da_buff+2, _T("%x"),   &val32[i]) ; /* it is a hex word */
            if (inival->valuetype == REG_QWORD) _stscanf_s(da_buff+2, _T("%llx"), &val64[i]) ; /* it is a hex word */
         }
         else
         {
            if (inival->valuetype == REG_DWORD) _stscanf_s(da_buff, _T("%u"),   &val32[i]) ;
            if (inival->valuetype == REG_QWORD) _stscanf_s(da_buff, _T("%llu"), &val64[i]) ;
         }
         da_buff = _tcstok_s(NULL, _T(",.; "), &next_token) ;
      }
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       generate_arraystring                                                                                            |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Generate a character string from an array of DWORDs or QWORDS                                                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void generate_arraystring(LPTSTR strng, size_t strng_size, INIVALUE *inival)
{
uint32_t i, numchar ;
int32_t j ;
LPTSTR format ;
   numchar = 0 ;
   strng[0] = _T('\0') ;
   format = _T("%u") ; /* default */
   if ((inival->valuetype == REG_DWORD) && ((inival->flags & REGFLAG_HEX) != 0)) format = _T("0x%x") ;
   if ((inival->valuetype == REG_QWORD) && ((inival->flags & REGFLAG_HEX) != 0)) format = _T("0x%llx") ;
   if ((inival->valuetype == REG_QWORD) && ((inival->flags & REGFLAG_HEX) == 0)) format = _T("%llu") ;
   for (i=0; i<inival->valuesize; i++)
   {
      if (i > 0)
      {
         strng[numchar] = _T(',') ;
         strng[numchar+1] = _T('\0') ;
         numchar++ ;
      }
      j = -1 ;
      if (inival->valuetype == REG_DWORD) j = _stprintf_s(&strng[numchar], strng_size-numchar, format, ((uint32_t *) inival->value)[i]) ;
      if (inival->valuetype == REG_QWORD) j = _stprintf_s(&strng[numchar], strng_size-numchar, format, ((uint64_t *) inival->value)[i]) ;
      if (j != -1) numchar += j ;
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       str2len                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Calculate the number of characters in an array of null-terminated strings, terminated by two null characters          |
| (REG_MULTI_SZ). The null characters between the individual strings are counted, but the terminating second null       |
| character is not counted similat to strlen.                                                                           |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint32_t str2len(LPTSTR str)
{
uint32_t i ;
   i = 0 ;
   while ((str[i] != _T('\0')) || (str[i+1] != _T('\0'))) i++ ;
   return(i+1) ; /* note that the first terminating null character is counted */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       clear_registry                                                                                                  |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Clear all registry keys which this program has created.                                                               |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void clear_registry(MYINI *myini)
{
HKEY keyhandle ;
LONG retv ;
uint32_t i ;
size_t keynamesize ;
TCHAR local_strng[MAX_REG_KEY_LENGHT] ;

   keynamesize = _tcsclen(myini->inikeyname) ;
   retv = RegOpenKeyEx(myini->hkey,
          myini->inikeyname,
          0,
          KEY_WRITE,
          &keyhandle) ;
   if (retv == ERROR_SUCCESS)
   { /* Note that keys can only be deleted, if they have no subkeys. Therefore all subkeys must be removed first. */
      for (i=0; i<myini->size; i++)
      {
         _tcscpy_s(local_strng, _countof(local_strng), myini->inikeyname) ;
         _tcscat_s(local_strng, _countof(local_strng)-keynamesize, _T("\\")) ;
         _tcscat_s(local_strng, _countof(local_strng)-keynamesize-1, myini->inivalues[i].sectionname) ;
         RegDeleteKey(myini->hkey, local_strng) ;
      }
      RegDeleteKey(myini->hkey, myini->inikeyname) ;
      RegCloseKey(keyhandle) ;
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void ShowError(int dialog)                                                                                      |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Display last error message                                                                                            |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void ShowError(int dialog)
{
LPTSTR lpMsgBuf ;

FormatMessage(
   FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
   NULL,
   GetLastError(),
   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
   (LPTSTR) &lpMsgBuf,
   0,
   NULL) ;
#ifdef _CONSOLE
   UNREFERENCED_PARAMETER(dialog) ;
   printf("%s", lpMsgBuf) ;
#else
   SetDlgItemText(MainWindow, dialog, lpMsgBuf) ;
#endif
// Free the buffer.
   LocalFree(lpMsgBuf) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void ShowSystemError(int errnr, int dialog)                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Display system error message                                                                                          |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void ShowSystemError(int errnr, int dialog)
{
LPTSTR lpMsgBuf ;

FormatMessage(
   FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
   NULL,
   errnr,
   MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), // Default language
   (LPTSTR) &lpMsgBuf,
   0,
   NULL) ;
#ifdef _CONSOLE
   UNREFERENCED_PARAMETER(dialog) ;
   printf("%s", lpMsgBuf) ;
#else
   SetDlgItemText(MainWindow, dialog, lpMsgBuf) ;
#endif
// Free the buffer.
   LocalFree(lpMsgBuf) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       set_process_prio                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Set priority of current process.                                                                                      |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void set_process_prio(int32_t prio)
{
int32_t last = (sizeof(process_prio)/sizeof(process_prio[0])) - 1 ;
   if (prio < 0) prio = 0 ;
   if (prio > last) prio = last ;
   SetPriorityClass(GetCurrentProcess(), process_prio[prio]) ; /* set priority */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       set_thread_prio                                                                                                 |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Set priority of a thread.                                                                                             |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void set_thread_prio(HANDLE hthread, int32_t prio)
{
int32_t last = (sizeof(thread_prio)/sizeof(thread_prio[0])) - 1 ;
   if (prio < 0) prio = 0 ;
   if (prio > last) prio = last ;
   SetThreadPriority(hthread, thread_prio[prio]) ; /* set priority */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       void display_info(char *msg)                                                                                   |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Display messages on the DLG_INFO control. DLG_INFO is a multi line edit control.                                      |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
//void display_info(char *msg)
//{
//HWND msg_hwnd ;
//uint32_t iLen ;
//   msg_hwnd = GetDlgItem(MainWindow, DLG_INFO) ;
//   iLen = GetWindowTextLength(msg_hwnd) ;
//   SendMessage(msg_hwnd, EM_SETSEL, iLen, iLen);
//   SendMessage(msg_hwnd, EM_REPLACESEL, (BOOL) false, (WPARAM) msg) ;
//   iLen = GetWindowTextLength(msg_hwnd) ;
//   SendMessage(msg_hwnd, EM_SETSEL, iLen, iLen);
//   SendMessage(msg_hwnd, EM_REPLACESEL, (BOOL) false, (WPARAM) NEWLINE) ;
//}


